use std::cmp::min;
use std::error::Error;
use std::fs::File;
use std::io::{BufRead, BufReader};
use std::mem::swap;
use std::path::Path;
use std::time::Instant;

const PATH_QUERY: &str = "./query.txt";
const PATH_DATASET: &str = "./dataset.txt";


fn main() {
    let query = read_lines(PATH_QUERY, 100).unwrap();
    let dataset = read_lines(PATH_DATASET, 100_000).unwrap();

    let args: Vec<String> = std::env::args().collect();
    let mut mode = "bruteforce";
    if args.len() == 2 {
        mode = &args[1];
    }

    let start = Instant::now();
    match mode {
        "bruteforce" => bruteforce(&query, &dataset),
        _ => panic!("Unsupported mode: {}", mode),
    }
    let elapsed = start.elapsed();
    println!("Elapsed: {}", elapsed.as_millis());
}

fn read_lines<P: AsRef<Path>>(path: P, estimated_capacity: usize) -> Result<Vec<String>, Box<dyn Error>> {
    let file = File::open(path)?;
    let mut reader = BufReader::with_capacity(1024 * 16, file);
    let mut data = Vec::with_capacity(estimated_capacity);
    let mut buffer = std::string::String::with_capacity(512);

    while reader.read_line(&mut buffer)? > 0 {
        let line = buffer.trim();
        data.push(line.to_owned());
        buffer.clear();
    }

    Ok(data)
}

#[inline(never)]
pub fn levenshtein<T>(a: &[T], b: &[T], max_distance: usize) -> usize
    where T: Eq + Copy
{
    // if a.len() == 0 {
    //     if b.len() > max_distance {
    //         return max_distance;
    //     }
    //     return b.len();
    // }

    let mut a = a;
    let mut b = b;
    if a.len() < b.len() {
        swap(&mut a, &mut b);
    }

    let ai = a.iter();
    let bi = b.iter();
    let prefix_len = common_prefix(ai, bi);

    let ai = a.iter().skip(prefix_len).rev();
    let bi = b.iter().skip(prefix_len).rev();
    let suffix_len = common_prefix(ai, bi);

    a = &a[prefix_len..a.len() - suffix_len];
    b = &b[prefix_len..b.len() - suffix_len];

    let cache_size = min(a.len(), b.len());
    let mut cache: Vec<usize> = (1..=cache_size).collect();
    let mut result = 0;

    for (i, &a_elem) in a.iter().enumerate() {
        result = i + 1;
        let mut dist_b = i;
        let mut current_distance = i;

        for (j, &b_elem) in b.iter().enumerate() {
            let dist_a = if a_elem == b_elem {
                dist_b
            } else {
                dist_b + 1
            };

            dist_b = cache[j];
            result = min(
                result + 1,
                min(dist_a, dist_b + 1),
            );
            cache[j] = result;

            if result < current_distance {
                current_distance = result;
            }
        }

        if current_distance >= max_distance {
            return max_distance;
        }
    }

    result
}

#[inline]
fn common_prefix<'b, X, I>(mut a: I, mut b: I) -> usize
    where
        I: Iterator<Item=&'b X>,
        X: Copy + PartialEq<X> + 'b,
{
    let mut prefix = 0;
    while let Some(&x) = b.next() {
        if let Some(&y) = a.next() {
            if x != y {
                break;
            }

            prefix += 1;
            continue;
        }
        break;
    }

    prefix
}

fn bruteforce<'a>(query: &'a [String], dataset: &'a [String]) {
    for (row_num, q) in query.iter().enumerate()
    {
        let mut min_distance = usize::MAX;
        //let mut nearest = None;

        let start = Instant::now();
        for entry in dataset.iter() {
            let distance = levenshtein(q.as_bytes(), entry.as_bytes(), min_distance);

            if distance < min_distance {
                min_distance = distance;
                //nearest = Some(entry);
            }
        }

        println!("{:3}|> Elapsed: {:5}; Distance: {:3}", row_num, start.elapsed().as_millis(), min_distance);
    }
}
